#! /usr/bin/perl -w

#
# here we link libraries
#
use strict;
use warnings;
#use SOptions;
use Getopt::Long;
# use ShowStruct;
use CIFParser;

#
# here we define variables
#	first - constant values
#	then  - common variables
#
my $version = 0.1;

my $CIFfile;
my $CIFtags;
my $outputFile;
my $scriptOutput;
my $dictFile;
my $dictTags;
my $quiet = 0;
my $parser = new CIFParser;

#
# here we declare subroutines
#

# subroutine to print module version and copyright notice
sub VersionMessage;

# subroutine to print help on module usage
sub HelpMessage;

# subroutine to extract tags from CIFfile array
sub getTags;

# subroutine to extract tags from data block array
sub getTagsSData;

# subroutine to convert reference (r. to array, hash, scalar) to scalar
sub refToScalar;

# subroutine to extract tags from dictionary (parsed using CIFParser)
sub getDict;

# subroutine to check if tags in data block provided are defined within 
# dictionary. It takes arguments:
#	- reference to data block array
#	- data block index (of array @{$CIFfile})
#	- data file index (of array @ARGV)
sub checkTags;

# subroutine to end script execution
# optionally takes single parameter - error message (scalar)
sub scriptExit;

#
# main program code
#

# check parameters passed
Getopt::Long::GetOptions
	("output=s"			=> \$outputFile,
		"dictionary=s"	=> \$dictFile,
		"version"		=> sub { VersionMessage() },
		"quiet"			=> sub { $quiet = 1; },
		"no-quiet"		=> sub { $quiet = 0; },
		"help|?"		=> sub { HelpMessage() }
	);

if( defined( $outputFile ) )
{
    open scriptOutputFH, ">$outputFile";
    $scriptOutput = \*scriptOutputFH;
} else {
    $scriptOutput = \*STDOUT;
}
print $scriptOutput; exit();

if( !defined($dictFile) )
{
	$dictFile = "cif_core.dic";
}
open( CIFdict, "<", $dictFile ) or scriptExit "Unable to open .
    $dictFile for reading.";
my $dt = getDict( $dictFile = $parser->Run( \*$scriptOutput, 
                            \*::CIFdict ) );
close( CIFdict );
%{$dictTags} = map { lc($_) => $_ } @{$dt};

if($quiet == 0)
{
	print $scriptOutput "Got dictionary tags:\n";
	foreach my $key ( keys %{$dictTags} )
	{
	    print $scriptOutput $key . "\t-->\t" . $dictTags->{$key} . "\n";
    }
	print $scriptOutput "\n";
}

# parse CIF files. If none is passed - display help message
if(@ARGV > 0)
{
	if(@ARGV > 1)
	{ # we have multiple files
		my $filen = 0;
		while($filen < @ARGV)
		{ # iterate through files
		    open( CIF, "<", $ARGV[$filen] ) or scriptExit "Unable to .
		                open $ARGV[$filen] for reading.";
			$CIFtags = getTags($parser->Run( \*$scriptOutput, \*CIF ));
			close( CIF );
			if( $quiet == 0 )
			{
				print $scriptOutput "Got following tags from file:\n";
				# showRef($CIFtags);
				print $scriptOutput "\n";
			}
			if( scalar @{$CIFtags} > 1 )
			{ # more than one data block
				for( my $i = 0; $i < ( scalar (@{$CIFtags}) ); $i++)
				{ # iterate through data block
					my $data = @{$CIFtags}[$i]->{content};
					checkTags($data, $i, $filen);
				}
			} else { # single data block
				my $data = @{$CIFtags}[0]->{content};
				checkTags($data, 0, $filen);
			}
			$filen++;
		}
	} else { # we have single file
	    open( CIF, "<", $ARGV[0] ) or scriptExit "Unable to .
	                open $ARGV[0] for reading.";
		$CIFtags = getTags($parser->Run( \*$scriptOutput, \*CIF));
		close(CIF);
		if( scalar @{$CIFtags} > 1 )
		{ # more than one data block
			for( my $i = 0; $i < ( scalar (@{$CIFtags}) ); $i++)
			{ # iterate through data block
				my $data = @{$CIFtags}[$i]->{content};
				checkTags($data, $i, 0);
			}
		} else { # single data block
			my $data = @{$CIFtags}[0]->{content};
			checkTags($data, 0, 0);
		}
	}
} else {
	HelpMessage();
	exit();
}

#
# here goes all subroutines bodies
#

sub VersionMessage
{
	print <<END_M;
cif_check_dict (using CIFParser v.$CIFParser::version) v.$version
Copyright (C) 2007 ???
This is free software. You may redistribute copies of it under the terms of
the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.
There is NO WARRANTY, to the extent permitted by law.

Written by Justas Butkus.
END_M
    exit(0);
}

sub HelpMessage
{
	print <<END_M;
usage: cif_check_dict [--output=<file_name>] [--dictionary=<file_name>] 
	[--version] [--quiet|--no-quiet] [--help] <cif_file1> 
	[,<cif_file2>, ..., <cif_fileN>]

If you specify OUTPUT, script will output anything to file given. 
Otherwise output will be sent to your screen (stdout).
	
If you specify DICTIONARY file - script will check CIF file provided 
  (look further) against that dictionary file (it itself MUST BE a 
  valid CIF dictionary). Otherwise it is attempted to check against 
  default dictionary, packaged with script.
	
If --version is requested - version message is printed and script 
  exits.
	
If NO-QUIET is specified, script will output additional information 
  - not only error mesages related to CIF file. This is the default 
  behaviour.
Otherwise, if QUIET is specified, script will output only error 
  messages, or warnings.
Don't specify both - you will only get a NO-QUIET behaviour...
	
If you type in --help - you get here.
	
And then you can add any number of CIF files that you would like to 
  check against dictionary.
END_M
    exit(0);
}

sub getTags
{
	my $file = shift;
	my $size = scalar @$file;
	my $tags;
	if($size > 1)
	{
		my $datan = 0;
		while($datan < $size)
		{
			push( @{$tags}, getTagsSData( $file->[$datan] ) );
			$datan++;
		}
	} else {
		$tags = [getTagsSData($file->[0])];
	}
	return $tags;
}

sub getTagsSData
{
	my $data = shift;
	my @tags;
	my $dataname = $data->{name};
	my $noitems = scalar @{$data->{content}};
	my $content = $data->{content};
	for(my $i = 0; $i < $noitems; $i++)
	{
		my $item = $content->[$i];
		if( $item->{kind} eq 'TAG' || $item->{kind} eq 'LOCAL' )
		{
			push(@tags, $item->{name});
		} elsif($item->{kind} eq 'loop') {
			foreach my $tag (@{$item->{name}})
			{
				push( @tags, $tag );
			}
		} elsif($item->{kind} eq 'SAVE') {
			my $savetags = getTagsSData( $item );
			foreach my $tag ( @{$savetags->{content}} )
			{
				push(@tags, $tag);
			}
		} else {
			die("ERROR: file contains elements, not handled by ".
			"this module!\n");
		}
	}
	return { kind => 'DATATAGS',
				name => $dataname,
				content => \@tags };
}

sub refToScalar
{
	my $reference = shift;
	my $indent;
	my $value = '';
	my $type = ref($reference);
	if( !ref($reference) )
	{
		return $reference;
	}
	if( $type eq "ARRAY" )
	{
		foreach my $entry ( @{$reference} )
		{
			if( !ref($entry) )
			{
				$value .= $entry . " | ";
			} else {
				$value .= refToScalar($entry)
					. "\n";
			}
		}
	} elsif ( $type eq "HASH" )
	{
		foreach my $key ( keys %{$reference} )
		{
			if( !ref($reference->{$key}) )
			{
				$value .= $reference->{$key} . " \\ ";
			} else {
				$value .= refToScalar($reference->{$key})
					. "\n";
			}
		}
	} elsif ( $type eq "SCALAR" ) {
		$value .= $$reference;
	} else {
		return $value;
	}
	return $value;
}

sub getDict
{
	my $dictF = shift;
	my $size = scalar @$dictF;
	my $tags;
	my $datan = 0;
	while($datan < $size)
	{
		my $convert = getDTagsSData( $dictF->[$datan] );
		if( $convert )
		{
			push( @{$tags}, $convert );
		}
		$datan++;
	}
	return $tags;
}

sub getDTagsSData
{
	my $data = shift;
	if( !($data->{name} eq "on_this_dictionary") )
	{
		my $tag;
		my $dataname = $data->{name};
		my $noitems = scalar @{$data->{content}};
		my $content = $data->{content};
		my $noreturn = 0;
		for(my $i = 0; $i < $noitems; $i++)
		{
			my $item = $content->[$i];
			if( $item->{kind} eq 'TAG' || $item->{kind} eq 'LOCAL' )
			{
				if( substr($item->{value},1) eq $dataname )
				{
					$tag = $item->{value};
				} elsif ( $item->{name} eq '_type' &&
							$item->{value} eq 'null' ) {
					$noreturn = 1;
				}
			} elsif($item->{kind} eq 'loop') {
				return 0;
			} elsif($item->{kind} eq 'SAVE') {
				my $savetags = getDTagsSData( $item );
				foreach my $tag ( @{$savetags->{content}} )
				{
					# check if not 0; in this case we got tag
					# which we do not want
					if( $tag )
					{
						if( substr($item->{value},1) eq $dataname )
						{
							$tag = $item->{value};
							print $scriptOutput $item->{value} . "\t\t" . $dataname . "\n";
						} elsif ( $item->{name} eq '_type' &&
									$item->{value} eq 'null' ) {
							print $scriptOutput $dataname . "\n";
							$noreturn = 1;
						}
					}
				}
			} else {
				die("ERROR: file contains elements, not handled by ".
				"this module!\n");
			}
		}
		if( $noreturn )
		{
			return 0;
		} else {
			return $tag;
		}
	} else {
		return 0;
	}
	return 0;
}

sub checkTags
{
	my $data = shift;
	my $i = shift;
	my $filen = shift;
	my %lcCIFtags;
	%lcCIFtags = map { lc($_) => $_ } @{$data};
	if( $quiet == 0 )
	{
		print $scriptOutput "Got following tags for file "
			. $ARGV[$filen] . ", data block "
			. @{$CIFtags}[$i]->{name} . ":\n";
		foreach my $key ( keys %lcCIFtags )
		{
		    print $scriptOutput $key . "\t-->\t"
		    		. $lcCIFtags{$key}
		    		. "\n";
	    }
	    print $scriptOutput "\n";
	}
	foreach my $key ( keys %lcCIFtags )
	{
		if( !defined $dictTags->{$key} )
		{
			print $scriptOutput <<END_M;
Got tag '$key' which is not in dictionary provided. This tag was found 
in file '$ARGV[$filen]', data block '@{$CIFtags}[$i]->{name}'.
Please check this tag - it might be wrong.

END_M
		}
	}
}


sub scriptExit
{
    if( defined( $outputFile ) )
    {
        close scriptOutputFH;
    }
    if( $#_ == -1  )
    {
        print STDERR $_[0];
        exit(1);
    }
    exit(0);
}
